@@ -11,15 +11,10 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */
 
 #include <pcu_l1_if.h>
 #include <gprs_rlcmac.h>
-#include <gsm_timer.h>
 #include <gprs_debug.h>
 #include <unistd.h>
 #include <stdbool.h>
@@ -31,18 +26,22 @@
 #include <signal.h>
 #include <sched.h>
 #include <bts.h>
-#include <gprs_coding_scheme.h>
 #include <osmocom/pcu/pcuif_proto.h>
 #include "gprs_bssgp_pcu.h"
+#include "alloc_algo.h"
 
 extern "C" {
 #include "pcu_vty.h"
+#include "coding_scheme.h"
+#include "pcu_l1_if_phy.h"
 #include <osmocom/gprs/gprs_bssgp.h>
-#include <osmocom/gprs/gprs_ns.h>
+#include <osmocom/gprs/gprs_ns2.h>
 #include <osmocom/vty/telnet_interface.h>
 #include <osmocom/vty/command.h>
 #include <osmocom/vty/vty.h>
 #include <osmocom/vty/ports.h>
+#include <osmocom/vty/logging.h>
+#include <osmocom/vty/cpu_sched_vty.h>
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/select.h>
 #include <osmocom/core/application.h>
@@ -63,32 +62,54 @@ extern void *bv_tall_ctx;
 static int quit = 0;
 static int rt_prio = -1;
 static bool daemonize = false;
-static const char *gsmtap_addr = "localhost"; // FIXME: use gengetopt's default value instead
+static const char *gsmtap_addr;
 
 static void print_help()
 {
-	printf( "Options:\n"
-		"  -h	--help		This text\n"
-		"  -c	--config-file 	Specify the filename of the config "
-			"file\n"
-		"  -m	--mcc MCC	Use given MCC instead of value "
-			"provided by BTS\n"
-		"  -n	--mnc MNC	Use given MNC instead of value "
-			"provided by BTS\n"
-		"  -V	--version	Print version\n"
-		"  -r	--realtime PRIO Use SCHED_RR with the specified "
-			"priority\n"
-		"  -D	--daemonize	Fork the process into a background "
-			"daemon\n"
-		"  -i	--gsmtap-ip	The destination IP used for GSMTAP\n"
+	printf( "Some useful options:\n"
+		"  -h	--help			This text\n"
+		"  -c	--config-file		Specify the filename of the config file\n"
+		"  -m	--mcc MCC		Use given MCC instead of value provided by BTS\n"
+		"  -n	--mnc MNC		Use given MNC instead of value provided by BTS\n"
+		"  -V	--version		Print version\n"
+		"  -D	--daemonize		Fork the process into a background daemon\n"
+		"\nVTY reference generation:\n"
+		"    	--vty-ref-mode MODE	VTY reference generation mode (e.g. 'expert').\n"
+		"    	--vty-ref-xml		Generate the VTY reference XML output and exit.\n"
 		);
 }
 
+static void handle_long_options(const char *prog_name, const int long_option)
+{
+	static int vty_ref_mode = VTY_REF_GEN_MODE_DEFAULT;
+
+	switch (long_option) {
+	case 1:
+		vty_ref_mode = get_string_value(vty_ref_gen_mode_names, optarg);
+		if (vty_ref_mode < 0) {
+			fprintf(stderr, "%s: Unknown VTY reference generation "
+				"mode '%s'\n", prog_name, optarg);
+			exit(2);
+		}
+		break;
+	case 2:
+		fprintf(stderr, "Generating the VTY reference in mode '%s' (%s)\n",
+			get_value_string(vty_ref_gen_mode_names, vty_ref_mode),
+			get_value_string(vty_ref_gen_mode_desc, vty_ref_mode));
+		vty_dump_xml_ref_mode(stdout, (enum vty_ref_gen_mode) vty_ref_mode);
+		exit(0);
+	default:
+		fprintf(stderr, "%s: error parsing cmdline options\n", prog_name);
+		exit(2);
+	}
+}
+
 /* FIXME: finally get some option parsing code into libosmocore */
 static void handle_options(int argc, char **argv)
 {
 	while (1) {
 		int option_idx = 0, c;
+		static int long_option = 0;
 		static const struct option long_options[] = {
 			{ "help", 0, 0, 'h' },
 			{ "config-file", 1, 0, 'c' },
@@ -99,6 +120,8 @@ static void handle_options(int argc, char **argv)
 			{ "daemonize", 0, 0, 'D' },
 			{ "exit", 0, 0, 'e' },
 			{ "gsmtap-ip", 1, 0, 'i' },
+			{ "vty-ref-mode", 1, &long_option, 1 },
+			{ "vty-ref-xml", 0, &long_option, 2 },
 			{ 0, 0, 0, 0 }
 		};
 
@@ -112,6 +135,9 @@ static void handle_options(int argc, char **argv)
 			print_help();
 			exit(0);
 			break;
+		case 0:
+			handle_long_options(argv[0], long_option);
+			break;
 		case 'c':
 			free(config_file);
 			config_file = strdup(optarg);
@@ -132,9 +158,13 @@ static void handle_options(int argc, char **argv)
 			break;
 		case 'i':
 			gsmtap_addr = optarg;
+			fprintf(stderr, "Command line argument '-i' is deprecated, use VTY "
+				"parameter 'gsmtap-remote-host %s' instead.\n", gsmtap_addr);
 			break;
 		case 'r':
 			rt_prio = atoi(optarg);
+			fprintf(stderr, "Command line argument '-r' is deprecated, use VTY "
+				"cpu-sched node setting 'policy rr %d' instead.\n", rt_prio);
 			break;
 		case 'D':
 			daemonize = true;
@@ -173,9 +203,16 @@ void sighandler(int sigset)
 		quit = 1;
 		break;
 	case SIGABRT:
-		/* in case of abort, we want to obtain a talloc report
-		 * and then return to the caller, who will abort the process
+		/* in case of abort, we want to obtain a talloc report and
+		 * then run default SIGABRT handler, who will generate coredump
+		 * and abort the process. abort() should do this for us after we
+		 * return, but program wouldn't exit if an external SIGABRT is
+		 * received.
 		 */
+		talloc_report_full(tall_pcu_ctx, stderr);
+		signal(SIGABRT, SIG_DFL);
+		raise(SIGABRT);
+		break;
 	case SIGUSR1:
 	case SIGUSR2:
 		talloc_report_full(tall_pcu_ctx, stderr);
@@ -186,7 +223,7 @@ void sighandler(int sigset)
 int main(int argc, char *argv[])
 {
 	struct sched_param param;
-	struct gprs_rlcmac_bts *bts;
+	struct gprs_pcu *pcu;
 	int rc;
 
 	/* tall_pcu_ctx may already have been initialized in bts.cpp during early_init(). */
@@ -197,80 +234,28 @@ int main(int argc, char *argv[])
 		osmo_init_logging2(tall_pcu_ctx, &gprs_log_info);
 	}
 
-	bts = bts_main_data();
-	bts->fc_interval = 1;
-	bts->initial_cs_dl = bts->initial_cs_ul = 1;
-	bts->initial_mcs_dl = bts->initial_mcs_ul = 1;
-	bts->cs1 = 1;
-	bts->n3101 = 10;
-	bts->n3103 = 4;
-	bts->n3105 = 8;
-	bts->alpha = 0; /* a = 0.0 */
-	bts->si13_is_set = false;
-	bts->cs_adj_enabled = 1;
-	bts->cs_adj_upper_limit = 33; /* Decrease CS if the error rate is above */
-	bts->cs_adj_lower_limit = 10; /* Increase CS if the error rate is below */
-	bts->max_cs_ul = MAX_GPRS_CS;
-	bts->max_cs_dl = MAX_GPRS_CS;
-	bts->max_mcs_ul = MAX_EDGE_MCS;
-	bts->max_mcs_dl = MAX_EDGE_MCS;
-	/* CS-1 to CS-4 */
-	bts->cs_lqual_ranges[0].low = -256;
-	bts->cs_lqual_ranges[0].high = 6;
-	bts->cs_lqual_ranges[1].low = 5;
-	bts->cs_lqual_ranges[1].high = 8;
-	bts->cs_lqual_ranges[2].low = 7;
-	bts->cs_lqual_ranges[2].high = 13;
-	bts->cs_lqual_ranges[3].low = 12;
-	bts->cs_lqual_ranges[3].high = 256;
+	pcu = gprs_pcu_alloc(tall_pcu_ctx);
+	the_pcu = pcu; /* globally avaialable object */
 
-	/* MCS-1 to MCS-9 */
-	/* Default thresholds are referenced from literature */
-	/* Fig. 2.3, Chapter 2, Optimizing Wireless Communication Systems, Springer (2009) */
-	bts->mcs_lqual_ranges[0].low = -256;
-	bts->mcs_lqual_ranges[0].high = 6;
-	bts->mcs_lqual_ranges[1].low = 5;
-	bts->mcs_lqual_ranges[1].high = 8;
-	bts->mcs_lqual_ranges[2].low = 7;
-	bts->mcs_lqual_ranges[2].high = 13;
-	bts->mcs_lqual_ranges[3].low = 12;
-	bts->mcs_lqual_ranges[3].high = 15;
-	bts->mcs_lqual_ranges[4].low = 14;
-	bts->mcs_lqual_ranges[4].high = 17;
-	bts->mcs_lqual_ranges[5].low = 16;
-	bts->mcs_lqual_ranges[5].high = 18;
-	bts->mcs_lqual_ranges[6].low = 17;
-	bts->mcs_lqual_ranges[6].high = 20;
-	bts->mcs_lqual_ranges[7].low = 19;
-	bts->mcs_lqual_ranges[7].high = 24;
-	bts->mcs_lqual_ranges[8].low = 23;
-	bts->mcs_lqual_ranges[8].high = 256;
-	bts->cs_downgrade_threshold = 200;
-
-	/* TODO: increase them when CRBB decoding is implemented */
-	bts->ws_base = 64;
-	bts->ws_pdch = 0;
-
-	bts->llc_codel_interval_msec = LLC_CODEL_USE_DEFAULT;
-	bts->llc_idle_ack_csec = 10;
-
-	/*
-	 * By default resegmentation is supported in DL
-	 * can also be configured through VTY
-	 */
-	bts->dl_arq_type = EGPRS_ARQ1;
-
-	bts->pcu_sock_path = talloc_strdup(tall_pcu_ctx, PCU_SOCK_DEFAULT);
+	pcu->pcu_sock_path = talloc_strdup(tall_pcu_ctx, PCU_SOCK_DEFAULT);
 
+	osmo_fsm_log_addr(false);
 	msgb_talloc_ctx_init(tall_pcu_ctx, 0);
 
 	osmo_stats_init(tall_pcu_ctx);
 	rate_ctr_init(tall_pcu_ctx);
-	gprs_ns_set_log_ss(DNS);
-	bssgp_set_log_ss(DBSSGP);
 
+	pcu_vty_info.tall_ctx = tall_pcu_ctx;
 	vty_init(&pcu_vty_info);
 	pcu_vty_init();
+	osmo_cpu_sched_vty_init(tall_pcu_ctx);
+	logging_vty_add_deprecated_subsys(tall_pcu_ctx, "bssgp");
+
+#ifdef ENABLE_DIRECT_PHY
+	rc = l1if_init();
+	if (rc < 0)
+		return -EINVAL;
+#endif
 
 	handle_options(argc, argv);
 	if ((!!spoof_mcc) + (!!spoof_mnc) == 1) {
@@ -279,19 +264,13 @@ int main(int argc, char *argv[])
 		exit(0);
 	}
 
-	bts->gsmtap = gsmtap_source_init(gsmtap_addr, GSMTAP_UDP_PORT, 1);
-
-	if (bts->gsmtap)
-		gsmtap_source_add_sink(bts->gsmtap);
-	else
-		fprintf(stderr, "Failed to initialize GSMTAP for %s\n", gsmtap_addr);
-
-	bssgp_nsi = gprs_ns_instantiate(&gprs_bssgp_ns_cb, tall_pcu_ctx);
-	if (!bssgp_nsi) {
+	pcu->nsi = gprs_ns2_instantiate(tall_pcu_ctx, gprs_ns_prim_cb, NULL);
+	if (!pcu->nsi) {
 		LOGP(DBSSGP, LOGL_ERROR, "Failed to create NS instance\n");
 		exit(1);
 	}
-	gprs_ns_vty_init(bssgp_nsi);
+	bssgp_set_bssgp_callback(gprs_gp_send_cb, pcu->nsi);
+	gprs_ns2_vty_init(pcu->nsi);
 
 	rc = vty_read_config_file(config_file, NULL);
 	if (rc < 0 && config_given) {
@@ -303,15 +282,39 @@ int main(int argc, char *argv[])
 		fprintf(stderr, "No config file: '%s' Using default config.\n",
 			config_file);
 
-	rc = telnet_init_dynif(tall_pcu_ctx, NULL, vty_get_bind_addr(),
-			       OSMO_VTY_PORT_PCU);
+	/* Accept a GSMTAP host from VTY config, but a commandline option overrides that. */
+	if (gsmtap_addr) {
+		if (pcu->gsmtap_remote_host != NULL) {
+			LOGP(DLGLOBAL, LOGL_NOTICE,
+			     "Command line argument '-i %s' overrides "
+			     "'gsmtap-remote-host %s' from the config file\n",
+			     gsmtap_addr, pcu->gsmtap_remote_host);
+			talloc_free(pcu->gsmtap_remote_host);
+		}
+		pcu->gsmtap_remote_host = talloc_strdup(pcu, gsmtap_addr);
+	}
+
+	if (pcu->gsmtap_remote_host) {
+		LOGP(DLGLOBAL, LOGL_NOTICE,
+		     "Setting up GSMTAP Um forwarding to '%s:%u'\n",
+		     pcu->gsmtap_remote_host, GSMTAP_UDP_PORT);
+		pcu->gsmtap = gsmtap_source_init(pcu->gsmtap_remote_host,
+							GSMTAP_UDP_PORT, 1);
+		if (pcu->gsmtap == NULL) {
+			fprintf(stderr, "Failed during gsmtap_source_init()\n");
+			exit(1);
+		}
+		gsmtap_source_add_sink(pcu->gsmtap);
+	}
+
+	rc = telnet_init_default(tall_pcu_ctx, NULL, OSMO_VTY_PORT_PCU);
 	if (rc < 0) {
 		fprintf(stderr, "Error initializing telnet\n");
 		exit(1);
 	}
 
-	if (!bts->alloc_algorithm)
-		bts->alloc_algorithm = alloc_algorithm_dynamic;
+	if (!pcu->alloc_algorithm)
+		pcu->alloc_algorithm = alloc_algorithm_dynamic;
 
 	rc = pcu_l1if_open();
 
@@ -347,10 +350,6 @@ int main(int argc, char *argv[])
 	}
 
 	while (!quit) {
-		osmo_gsm_timers_check();
-		osmo_gsm_timers_prepare();
-		osmo_gsm_timers_update();
-
 		osmo_select_main(0);
 	}
 
@@ -358,7 +357,7 @@ int main(int argc, char *argv[])
 
 	pcu_l1if_close();
 
-	bts_cleanup();
+	TALLOC_FREE(the_pcu);
 	talloc_report_full(tall_pcu_ctx, stderr);
 	talloc_free(tall_pcu_ctx);
 
