@@ -20,7 +20,6 @@
 #include <stdlib.h>
 #include "pcu_vty_functions.h"
 #include "bts.h"
-#include "gprs_ms_storage.h"
 #include "gprs_ms.h"
 #include "cxx_linuxlist.h"
 #include <llc.h>
@@ -40,29 +39,41 @@ extern "C" {
 	#include "coding_scheme.h"
 }
 
+static uint32_t tbf_state_flags(const struct gprs_rlcmac_tbf *tbf)
+{
+	const struct gprs_rlcmac_ul_tbf *ul_tbf = tbf_as_ul_tbf_const(tbf);
+	const struct gprs_rlcmac_dl_tbf *dl_tbf = tbf_as_dl_tbf_const(tbf);
+	if (ul_tbf)
+		return ul_tbf->state_fsm.state_flags;
+	return dl_tbf->state_fsm.state_flags;
+}
+
 static void tbf_print_vty_info(struct vty *vty, struct gprs_rlcmac_tbf *tbf)
 {
-	gprs_rlcmac_ul_tbf *ul_tbf = as_ul_tbf(tbf);
-	gprs_rlcmac_dl_tbf *dl_tbf = as_dl_tbf(tbf);
+	gprs_rlcmac_ul_tbf *ul_tbf = tbf_as_ul_tbf(tbf);
+	gprs_rlcmac_dl_tbf *dl_tbf = tbf_as_dl_tbf(tbf);
+	uint32_t state_flags = tbf_state_flags(tbf);
+	struct GprsMs *ms = tbf_ms(tbf);
+	const struct gprs_rlcmac_pdch *first_common_ts = ms_first_common_ts(ms);
 
 	vty_out(vty, "TBF: TFI=%d TLLI=0x%08x (%s) TA=%u DIR=%s IMSI=%s%s", tbf->tfi(),
 		tbf->tlli(), tbf->is_tlli_valid() ? "valid" : "invalid",
 		tbf->ta(),
 		tbf->direction == GPRS_RLCMAC_UL_TBF ? "UL" : "DL",
 		tbf->imsi(), VTY_NEWLINE);
-	vty_out(vty, " created=%lu state=%s flags=%08x [CCCH:%u, PACCH:%u] 1st_TS=%d 1st_cTS=%d ctrl_TS=%d MS_CLASS=%d/%d%s",
+	vty_out(vty, " created=%lu state=%s flags=%08x [CCCH:%u, PACCH:%u] 1st_cTS=%" PRId8 " ctrl_TS=%d MS_CLASS=%d/%d%s",
 		tbf->created_ts(), tbf->state_name(),
-		tbf->state_fsm.state_flags,
-		tbf->state_fsm.state_flags & (1 << GPRS_RLCMAC_FLAG_CCCH),
-		tbf->state_fsm.state_flags & (1 << GPRS_RLCMAC_FLAG_PACCH),
-		tbf->first_ts,
-		tbf->first_common_ts, tbf->control_ts,
+		state_flags,
+		state_flags & (1 << GPRS_RLCMAC_FLAG_CCCH),
+		state_flags & (1 << GPRS_RLCMAC_FLAG_PACCH),
+		first_common_ts ? first_common_ts->ts_no : -1,
+		tbf->control_ts ? tbf->control_ts->ts_no : -1,
 		tbf->ms_class(),
-		ms_egprs_ms_class(tbf->ms()),
+		ms_egprs_ms_class(ms),
 		VTY_NEWLINE);
 	vty_out(vty, " TS_alloc=");
 	for (int i = 0; i < 8; i++) {
-		bool is_ctrl = tbf->is_control_ts(i);
+		bool is_ctrl = tbf_is_control_ts(tbf, tbf->pdch[i]);
 		if (tbf->pdch[i])
 			vty_out(vty, "%d%s ", i, is_ctrl ? "!" : "");
 	}
@@ -101,6 +112,8 @@ int pcu_vty_show_tbf_all(struct vty *vty, struct gprs_rlcmac_bts *bts, uint32_t
 	struct llist_item *iter;
 	const struct gprs_rlcmac_trx *trx;
 	struct gprs_rlcmac_tbf *tbf;
+	const struct gprs_rlcmac_ul_tbf *ul_tbf;
+	const struct gprs_rlcmac_dl_tbf *dl_tbf;
 	size_t trx_no;
 
 	vty_out(vty, "UL TBFs%s", VTY_NEWLINE);
@@ -108,7 +121,8 @@ int pcu_vty_show_tbf_all(struct vty *vty, struct gprs_rlcmac_bts *bts, uint32_t
 		trx = &bts->trx[trx_no];
 		llist_for_each_entry(iter, &trx->ul_tbfs, list) {
 			tbf = (struct gprs_rlcmac_tbf *)iter->entry;
-			if (tbf->state_fsm.state_flags & flags)
+			ul_tbf = tbf_as_ul_tbf_const(tbf);
+			if (ul_tbf->state_fsm.state_flags & flags)
 				tbf_print_vty_info(vty, tbf);
 		}
 	}
@@ -118,7 +132,8 @@ int pcu_vty_show_tbf_all(struct vty *vty, struct gprs_rlcmac_bts *bts, uint32_t
 		trx = &bts->trx[trx_no];
 		llist_for_each_entry(iter, &trx->dl_tbfs, list) {
 			tbf = (struct gprs_rlcmac_tbf *)iter->entry;
-			if (tbf->state_fsm.state_flags & flags)
+			dl_tbf = tbf_as_dl_tbf_const(tbf);
+			if (dl_tbf->state_fsm.state_flags & flags)
 				tbf_print_vty_info(vty, tbf);
 		}
 	}
@@ -133,11 +148,16 @@ static int show_ms(struct vty *vty, GprsMs *ms)
 	uint8_t slots;
 
 	vty_out(vty, "MS TLLI=%08x, IMSI=%s%s", ms_tlli(ms), ms_imsi(ms), VTY_NEWLINE);
-	vty_out(vty, "  Timing advance (TA):    %d%s", ms_ta(ms), VTY_NEWLINE);
-	vty_out(vty, "  Coding scheme uplink:   %s%s", mcs_name(ms_current_cs_ul(ms)),
-		VTY_NEWLINE);
-	vty_out(vty, "  Coding scheme downlink: %s%s", mcs_name(ms_current_cs_dl(ms, ms_mode(ms))),
-		VTY_NEWLINE);
+	if (osmo_timer_pending(&ms->release_timer)) {
+		struct timeval tv_now, tv_res1, tv_res2;
+		osmo_gettimeofday(&tv_now, NULL);
+		timersub(&tv_now, &ms->tv_idle_start, &tv_res1);
+		osmo_timer_remaining(&ms->release_timer, &tv_now, &tv_res2);
+		vty_out(vty, "  State:                  IDLE for %lus, release in %lus%s",
+			tv_res1.tv_sec, tv_res2.tv_sec, VTY_NEWLINE);
+	} else {
+		vty_out(vty, "  State:                  ACTIVE%s", VTY_NEWLINE);
+	}
 	vty_out(vty, "  Mode:                   %s%s", mode_name(ms_mode(ms)), VTY_NEWLINE);
 	vty_out(vty, "  MS class:               %d%s", ms_ms_class(ms), VTY_NEWLINE);
 	vty_out(vty, "  EGPRS MS class:         %d%s", ms_egprs_ms_class(ms), VTY_NEWLINE);
@@ -145,12 +165,17 @@ static int show_ms(struct vty *vty, GprsMs *ms)
 	slots = ms_current_pacch_slots(ms);
 	for (int i = 0; i < 8; i++)
 		if (slots & (1 << i))
-			vty_out(vty, "%d ", i);
+			vty_out(vty, "TS%d ", i);
 	vty_out(vty, "%s", VTY_NEWLINE);
-	vty_out(vty, "  LLC queue length:       %zd%s", llc_queue_size(ms_llc_queue(ms)),
+	vty_out(vty, "  DL LLC queue length:    %zd%s", llc_queue_size(ms_llc_queue(ms)),
 		VTY_NEWLINE);
-	vty_out(vty, "  LLC queue octets:       %zd%s", llc_queue_octets(ms_llc_queue(ms)),
+	vty_out(vty, "  DL LLC queue octets:    %zd%s", llc_queue_octets(ms_llc_queue(ms)),
 		VTY_NEWLINE);
+	vty_out(vty, "  DL Coding Scheme:       %s%s", mcs_name(ms_current_cs_dl(ms, ms_mode(ms))),
+		VTY_NEWLINE);
+	vty_out(vty, "  UL Coding Scheme:       %s%s", mcs_name(ms_current_cs_ul(ms)),
+		VTY_NEWLINE);
+	vty_out(vty, "  Timing advance (TA):    %d%s", ms_ta(ms), VTY_NEWLINE);
 	if (ms->l1_meas.have_rssi)
 		vty_out(vty, "  RSSI:                   %d dBm%s",
 			ms->l1_meas.rssi, VTY_NEWLINE);
@@ -167,7 +192,7 @@ static int show_ms(struct vty *vty, GprsMs *ms)
 		vty_out(vty, "  Downlink NACK rate:     %d %%%s",
 			ms_nack_rate_dl(ms), VTY_NEWLINE);
 	if (ms->l1_meas.have_ms_rx_qual)
-		vty_out(vty, "  MS RX quality:          %d %%%s",
+		vty_out(vty, "  MS Rx quality:          %d %%%s",
 			ms->l1_meas.ms_rx_qual, VTY_NEWLINE);
 	if (ms->l1_meas.have_ms_c_value)
 		vty_out(vty, "  MS C value:             %d dB%s",
@@ -181,25 +206,25 @@ static int show_ms(struct vty *vty, GprsMs *ms)
 				i, ms->l1_meas.ts[i].ms_i_level, VTY_NEWLINE);
 	}
 	if (ms_ul_tbf(ms))
-		vty_out(vty, "  Uplink TBF:             TFI=%d, state=%s%s",
+		vty_out(vty, "  UL TBF:                 TFI=%d, state=%s%s",
 			ms_ul_tbf(ms)->tfi(),
 			ms_ul_tbf(ms)->state_name(),
 			VTY_NEWLINE);
 	if (ms_dl_tbf(ms)) {
-		vty_out(vty, "  Downlink TBF:           TFI=%d, state=%s%s",
+		vty_out(vty, "  DL TBF:                 TFI=%d, state=%s%s",
 			ms_dl_tbf(ms)->tfi(),
 			ms_dl_tbf(ms)->state_name(),
 			VTY_NEWLINE);
-		vty_out(vty, "  Current DL Throughput:  %d Kbps %s",
+		vty_out(vty, "  Current DL Throughput:  %d Kbps%s",
 			ms_dl_tbf(ms)->m_bw.dl_throughput,
 			VTY_NEWLINE);
 	}
 
 	llist_for_each_entry(i_tbf, &ms->old_tbfs, list) {
 		struct gprs_rlcmac_tbf *tbf = (struct gprs_rlcmac_tbf *)i_tbf->entry;
-		vty_out(vty, "  Old %-19s TFI=%d, state=%s%s",
+		vty_out(vty, "  Old %s TBF: TFI=%d, state=%s%s",
 			tbf_direction(tbf) == GPRS_RLCMAC_UL_TBF ?
-			"Uplink TBF:" : "Downlink TBF:",
+			"UL" : "DL",
 			tbf->tfi(),
 			tbf->state_name(),
 			VTY_NEWLINE);
@@ -213,8 +238,8 @@ int pcu_vty_show_ms_all(struct vty *vty, struct gprs_rlcmac_bts *bts)
 {
 	struct llist_head *tmp;
 
-	llist_for_each(tmp, bts_ms_store(bts)->ms_list()) {
-		GprsMs *ms_iter = llist_entry(tmp, typeof(*ms_iter), list);
+	llist_for_each(tmp, &bts->ms_list) {
+		struct GprsMs *ms_iter = llist_entry(tmp, typeof(*ms_iter), list);
 		show_ms(vty, ms_iter);
 	}
 
@@ -224,7 +249,7 @@ int pcu_vty_show_ms_all(struct vty *vty, struct gprs_rlcmac_bts *bts)
 int pcu_vty_show_ms_by_tlli(struct vty *vty, struct gprs_rlcmac_bts *bts,
 	uint32_t tlli)
 {
-	GprsMs *ms = bts_ms_store(bts)->get_ms(tlli);
+	struct GprsMs *ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	if (!ms) {
 		vty_out(vty, "Unknown TLLI %08x.%s", tlli, VTY_NEWLINE);
 		return CMD_WARNING;
@@ -236,7 +261,7 @@ int pcu_vty_show_ms_by_tlli(struct vty *vty, struct gprs_rlcmac_bts *bts,
 int pcu_vty_show_ms_by_imsi(struct vty *vty, struct gprs_rlcmac_bts *bts,
 	const char *imsi)
 {
-	GprsMs *ms = bts_ms_store(bts)->get_ms(0, 0, imsi);
+	struct GprsMs *ms = bts_get_ms_by_imsi(bts, imsi);
 	if (!ms) {
 		vty_out(vty, "Unknown IMSI '%s'.%s", imsi, VTY_NEWLINE);
 		return CMD_WARNING;
@@ -264,7 +289,7 @@ int pcu_vty_show_bts_pdch(struct vty *vty, const struct gprs_rlcmac_bts *bts)
 		for (ts_nr = 0; ts_nr < ARRAY_SIZE(trx->pdch); ts_nr++) {
 			const struct gprs_rlcmac_pdch *pdch = &trx->pdch[ts_nr];
 
-			vty_out(vty, "  TS%u: PDCH %s, %u UL TBFs, %u DL TBFs%s", pdch->ts_no,
+			vty_out(vty, "  TS%u: PDCH %s, %u DL TBFs, %u UL TBFs%s", pdch->ts_no,
 				pdch->is_enabled() ? "enabled" : "disabled",
 				pdch->num_tbfs(GPRS_RLCMAC_DL_TBF),
 				pdch->num_tbfs(GPRS_RLCMAC_UL_TBF), VTY_NEWLINE);
